/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2009 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonab ly 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


#include <stdio.h>
#include <unistd.h>
#include <getopt.h>

#include "base/adapter/PlatformAdapter.h"
#include "SyncProcessManager.h"
#include "constants.h"

USE_NAMESPACE

void setupLog();

int main (int argc, char **argv) 
{
    int c = 0, i = 0;
    int option_index = 0;
    int verbose_flag = 0;
    int sync_contacts = 0;
    int sync_config = 0;
    int cleanup_contacts = 0;
    int use_wait_queue = 0;
    int sources_num = 0;
    int status = EXIT_SUCCESS;
    const char* sources_list[MAX_SYNC_SOURCES_NUM];
    
    sync_action_t syncType        = SYNC_TYPE_UNSET;
    sync_action_t currentSyncType = SYNC_TYPE_UNSET;

    struct option long_options[] = {
        {"sync-addressbook",    no_argument, 0, 'a' },
        {"sync-config",         no_argument, 0, 'c' },
        {"cleanup-contacts",    no_argument, 0, 'A' },
        {"verbose",             no_argument, 0, 'v' },
        {"help",                no_argument, 0, 'h' },
        {"queue",               no_argument, 0, 'W' },
        {0, 0, 0, 0}
    };
    
    NSMutableArray* sourcesNameList         = nil;
    NSAutoreleasePool *pool                 = nil;
    MacMappingStoreBuilder* builder         = NULL;
    NSThread* syncProcessManagerThread      = nil; // sync worker thread
    NSThread* sigHandlerThread              = nil; // sig handler thread
    SyncProcessManager *syncProcessManager  = nil;
    SigHandler *sigHandler                  = nil; 
    NSString*  sourceName                   = nil;
    
    StringBuffer opt;
    StringBuffer help;
    
    help.sprintf("%s MacOsX Sync Client\n\nUsage:\n--sync-addressbook [-a] to sync contacts\n--cleanup-addressbook [-A]"
                  " to reset your local contacts\n--sync-config [-c] to authenticate\n--queue [-W] to queue the process "
                  "while another sync is running (default behaviour is exiting with error).\n\n", APPNAME);
    
    if (argc < 2) {
        fprintf(stderr, "%s", help.c_str());
        exit(EXIT_FAILURE);
    } 
  
    while ((c = getopt_long(argc, argv, "acAvhW", long_options, &option_index)) != EOF) 
    {
        switch (c) {
            case 'a':
                if (!sync_contacts) {
                    if (cleanup_contacts) {
                        fprintf(stderr, "mismatching options selected");
                        exit(EXIT_FAILURE);
                    }
                    
                    sync_contacts = 1;
                    sources_list[sources_num++] = CONTACTS_SOURCE_NAME;
                    syncType = SYNC_CONTACTS;
                }
                
                break;
 
            case 'c':
                if (!sync_config) {
                    sync_config = 1;
                    sources_list[sources_num++] = CONFIG_SOURCE_NAME;
                    syncType = SYNC_CONFIG;
                }
                
                break;
 
            case 'A':
                if (cleanup_contacts == 0) {
                    if (sync_contacts) {
                        fprintf(stderr, "mismatching options selected\n");
                        exit(EXIT_FAILURE);
                    }
                    
                    cleanup_contacts = 1; 
                    sources_list[sources_num++] = CONTACTS_SOURCE_NAME;
                    syncType = RESET_CONTACTS;
                }
                
                break;
 
            case 'v':
                verbose_flag = 1;
                break;
           
            case 'h':
                fprintf(stdout, "%s", help.c_str());
                exit(EXIT_SUCCESS);
 
            case 'W':
                use_wait_queue = 1;
                break;

            case '?':
            default:
                fprintf(stderr, "%s", help.c_str());
                exit(EXIT_FAILURE);
         }
    }

    // initialize the PlatformAdapter and create a default config if needed
    ClientConfig::getInstance()->init();
   
    setupLog();
    
    if (sources_num == 0) {
        LOG.error("no valid source to sync specified: exiting");
        fprintf(stderr, "%s", help.c_str());
        
        exit(EXIT_FAILURE);
    }

    // check if there's sync info shared memory
    // and if its creator it's still running
    if (checkCurrentSync() == false) {
        // orphaned sync info shared mem: remove it
        LOG.info("found orphaned sync info shared memory: trying to remove it");
        if (removeCurrentSyncType()) {
            LOG.error("error removing shared memory: %s", strerror(errno));
            exit(EXIT_FAILURE);
        }
    }
    
    if (use_wait_queue) {
        currentSyncType = getCurrentSyncType();
        
        if (currentSyncType == syncType) {
            LOG.info("another sync for the same source is currently running: exiting from wait queue");
            exit(EXIT_SUCCESS);
        } else if (currentSyncType != SYNC_TYPE_UNSET) {
            if (waitForRunningProcesses(SYNC_APPLICATION_NAME) < 0) {
                LOG.error("error in wait queue");
                exit(EXIT_FAILURE);
            }
        }
    } else { 
        if (checkProcessRunning(SYNC_APPLICATION_NAME, 1)) {
            LOG.error("another sync is currently running: exiting...");
        
            exit(EXIT_FAILURE);
        }
    }

    if (setCurrentSyncType(syncType)) {
        exit(EXIT_FAILURE);
    }

    builder = new MacMappingStoreBuilder();
    MappingsManager::setBuilder(builder);
    
    pool = [[NSAutoreleasePool alloc] init];

    sourcesNameList = [[NSMutableArray alloc] initWithCapacity: MAX_SYNC_SOURCES_NUM];
    
    for (i = 0; i < sources_num; i++) {
        NSString *sourceName = [NSString stringWithUTF8String: sources_list[i] ];
        if ([sourceName length]) {
            [sourcesNameList addObject: sourceName];
        }
        [sourceName release];
    }

    syncProcessManager = [[SyncProcessManager alloc] init];
    
    if (cleanup_contacts) {
        syncProcessManagerThread = [[NSThread alloc] initWithTarget: syncProcessManager
                                                           selector: @selector(runResetThread:)
                                                             object: sourcesNameList];
        
    } else {
        syncProcessManagerThread = [[NSThread alloc] initWithTarget: syncProcessManager
                                                           selector: @selector(runSyncThread:)
                                                             object: sourcesNameList];
    }
    
    sigHandler = [[SigHandler alloc] init];
    
    sigHandlerThread = [[NSThread alloc] initWithTarget: sigHandler
                                                         selector: @selector(runSigHandler:)
                                                           object: (id)nil];
   
    // start signals handler thread
    [sigHandlerThread start];
    
    // start sync thread 
    [syncProcessManagerThread start];
    
    // this loop is terminated by the runThread method. Needed to get the info from the UI
    CFRunLoopRun();
        
    [syncProcessManagerThread dealloc];
    [syncProcessManager dealloc];
    [sourcesNameList dealloc];
    
    
    if (removeCurrentSyncType()) {
        LOG.error("error removing shared memory: %s", strerror(errno));
        status = EXIT_FAILURE;
    }

    // stop sig handler thread 
    [sigHandler stopSigHandler];
    [sigHandlerThread dealloc];
    [sigHandler dealloc];
   
    [pool drain];
   
    exit(status);
}

void setupLog()
{
    // initialize the PlatformAdapter and create a default config if needed
    StringBuffer logsFolder = PlatformAdapter::getConfigFolder();
    logsFolder += "/logs/";
    // FIXME - we should check this sys call
    mkdir(logsFolder, S_IRWXU);
    
    LOG.setLogPath(logsFolder);
    LOG.setLogName("/synclog.txt");
    LOG.setLevel(ClientConfig::getInstance()->getDeviceConfig().getLogLevel());
    
    time_t now = time(NULL);
    LOG.info("******************************************");
    LOG.info("Mac Sync Client: current UTC date and time");
    LOG.info("******************************************");
}

